import typing

import readline

from mccq.cli.meta import META_COMMANDS
from mccq.query_arguments import QueryArguments
from mccq.query_manager import QueryManager
from mccq.typedefs import IterableOfStrings


class CLICompleter:
    def __init__(self, query_manager: QueryManager):
        self.query_manager = query_manager
        self.completions: typing.List[str] = []
        self.num_completions: int = 0
        self.reset_completions()

    def make_meta_completions(self, line: str) -> IterableOfStrings:
        command = line[1:]
        yield from ('\\' + s for s in META_COMMANDS if s.startswith(command))

    def make_query_completions_for_version(self, version: str, arguments: QueryArguments) -> IterableOfStrings:
        # build a query tree as an intermediate result
        query_tree = self.query_manager.query_tree_for_version(version, arguments)

        # yield the key of the data node from each leaf of the query tree
        if query_tree:
            for leaf in query_tree.leaves():
                yield leaf.data_node.key

    def make_query_completions(self, line: str) -> IterableOfStrings:
        # this works similarly to actually querying a command, but instead of yielding all results we yield all leaf
        # nodes as suggestions for the next token

        # append wildcard to the end of the current command so we can invoke a mock query
        command = line + '.*'

        try:
            arguments = self.query_manager.parse_query_arguments(command)
            versions = self.query_manager.filter_versions(arguments)

        except:
            return None

        # spread across all versions applicable to the current command
        for v in versions:
            try:
                yield from self.make_query_completions_for_version(v, arguments)
            except:
                continue

    def make_completions(self, line: str) -> IterableOfStrings:
        if line.startswith('\\'):
            yield from self.make_meta_completions(line)

        else:
            yield from self.make_query_completions(line)

    def reset_completions(self):
        self.completions = []
        self.num_completions = 0

    def complete(self, text: str, state: int):
        if state:
            if state == self.num_completions:
                self.reset_completions()
                return

            return self.completions[state]

        else:
            line = readline.get_line_buffer()

            # append completions with a space so that singular results are completed in full
            self.completions = tuple(word + ' ' for word in set(self.make_completions(line)))

            self.num_completions = len(self.completions)

            return self.completions[0]



import json
import logging
import os

from mccq.data_loader.abc.data_loader import DataLoader
from mccq.typedefs import TupleOfStrings

log = logging.getLogger(__name__)


class FilesystemDataLoader(DataLoader):
    def load(self, components: TupleOfStrings) -> dict:
        path = os.path.join(*components)
        log.info(f'Loading commands from filesystem: {path}')
        with open(path) as fp:
            raw = json.load(fp)
        return raw

    def load_version(self, components: TupleOfStrings) -> str:
        path = os.path.join(*components)
        log.info(f'Loading version from filesystem: {path}')
        with open(path) as fp:
            raw = str(fp.readline()).strip()
        return raw



import abc
import typing


class Node(abc.ABC):
    def leaves(self) -> typing.Iterable['Node']:
        if self.children:
            for child in self.children:
                yield from child.leaves()
        else:
            yield self

    @property
    @abc.abstractmethod
    def children(self) -> typing.Tuple['Node', ...]: ...



from mccq.typedefs import TupleOfStrings


class QueryArguments:
    def __init__(
            self,
            command: TupleOfStrings,
            showtypes: bool = None,
            explode: bool = None,
            capacity: int = None,
            versions: TupleOfStrings = None,
    ):
        self.command = command
        self.showtypes = showtypes
        self.explode = explode
        self.capacity = capacity
        self.versions = versions



import argparse


class ArgumentParserError(Exception):
    pass


class ArgumentParserExit(Exception):
    pass


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)

    def exit(self, status=0, message=None):
        raise ArgumentParserExit(status, message)



import typing

from mccq.node.abc.node import Node


class DataNode(Node):
    def __init__(
            self,
            relevant: bool = None,
            population: int = None,
            key: str = None,
            command: str = None,
            command_t: str = None,
            argument: str = None,
            argument_t: str = None,
            collapsed: str = None,
            collapsed_t: str = None,
            children: typing.Tuple['DataNode', ...] = None,
    ):
        self.relevant = relevant
        self.population = population
        self.key = key
        self.command = command
        self.command_t = command_t
        self.argument = argument
        self.argument_t = argument_t
        self.collapsed = collapsed
        self.collapsed_t = collapsed_t
        self._children = children

    def leaves(self) -> typing.Iterable['DataNode']:
        return super().leaves()

    @property
    def children(self) -> typing.Tuple['DataNode', ...]:
        return self._children or ()



import logging
import os
import shlex
import sys

from mccq.argument_parser import ArgumentParser
from mccq.cli.loop import cli_loop
from mccq.query_manager import QueryManager
from mccq.version_database import VersionDatabase

# TODO other os, edge cases
local_database = os.path.join(os.path.expanduser('~'), 'AppData', 'Roaming', '.minecraft', 'versions')

startup_parser = ArgumentParser(
    'mccq',
    description='Minecraft command query program. Inspired by the in-game help command, with added features like '
                'version reporting and expandable regex search.')

startup_parser.add_argument(
    '-s', '--show_versions', action='append', default=[], help='which version(s) to render by default (repeatable)')

startup_parser.add_argument(
    '-d', '--database_uri', default=local_database, help='the uri from where versions will be loaded')

startup_parser.add_argument(
    '-l', '--log', default=logging.WARNING, help='log level')

try:
    startup_args = startup_parser.parse_args(shlex.split(' '.join(sys.argv[1:])))
except:
    sys.exit()

logging.basicConfig(level=startup_args.log)

log = logging.getLogger(__name__)

show_versions = startup_args.show_versions
database_uri = startup_args.database_uri

db = VersionDatabase(
    uri=database_uri)

qm = QueryManager(
    database=db,
    show_versions=show_versions)

print('[::] Minecraft Command Query CLI [::]')

# attempt to load extra features like navigation and autocompletion
try:
    import readline
    from mccq.cli.completer import CLICompleter

    cc = CLICompleter(qm)

    readline.set_completer_delims(' \t\n')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(cc.complete)

    print('Navigation and autocompletion are available on the current system.')

except:
    pass

print('Enter a Minecraft command query, or "exit" to leave.')

cli_loop(qm)

print('Goodbye!')



import abc

from mccq.typedefs import TupleOfStrings


class DataLoader(abc.ABC):
    @abc.abstractmethod
    def load(self, components: TupleOfStrings) -> dict: ...

    @abc.abstractmethod
    def load_version(self, components: TupleOfStrings) -> str: ...
